from jinja2 import Environment, meta
from traceloop.sdk.prompts.model import Prompt, PromptVersion, TemplateEngine
from traceloop.sdk.prompts.registry import PromptRegistry
from traceloop.sdk.tracing.tracing import set_prompt_tracing_context


def get_effective_version(prompt: Prompt) -> PromptVersion:
    if len(prompt.versions) == 0:
        raise Exception(f"No versions exist for {prompt.key} prompt")

    return next(v for v in prompt.versions if v.id == prompt.target.version)


class PromptRegistryClient:
    _registry: PromptRegistry
    _jinja_env: Environment

    def __new__(cls) -> "PromptRegistryClient":
        if not hasattr(cls, "instance"):
            obj = cls.instance = super(PromptRegistryClient, cls).__new__(cls)
            obj._registry = PromptRegistry()
            obj._jinja_env = Environment()

        return cls.instance

    def render_prompt(self, key: str, **args):
        prompt = self._registry.get_prompt_by_key(key)
        if prompt is None:
            raise Exception(f"Prompt {key} does not exist")

        prompt_version = get_effective_version(prompt)
        params_dict = {"messages": self.render_messages(prompt_version, **args)}
        params_dict.update(prompt_version.llm_config)
        params_dict.pop("mode")

        set_prompt_tracing_context(
            prompt.key, prompt_version.version, prompt_version.name, prompt_version.hash
        )

        return params_dict

    def render_messages(self, prompt_version: PromptVersion, **args):
        if prompt_version.templating_engine == TemplateEngine.JINJA2:
            rendered_messages = []
            for msg in prompt_version.messages:
                template = self._jinja_env.from_string(msg.template)
                template_variables = meta.find_undeclared_variables(
                    self._jinja_env.parse(msg.template)
                )
                missing_variables = template_variables.difference(set(args.keys()))
                if missing_variables == set():
                    rendered_msg = template.render(args)
                else:
                    raise Exception(
                        f"Input variables: {','.join(str(var) for var in missing_variables)} are missing"
                    )

                # TODO: support other types than openai chat structure
                rendered_messages.append({"role": msg.role, "content": rendered_msg})

            return rendered_messages
        else:
            raise Exception(
                f"Templating engine {prompt_version.templating_engine} is not supported"
            )



# Manages list of associated properties for which content tracing
# (prompts, vector embeddings, etc.) is allowed.
class ContentAllowList:
    def __new__(cls) -> "ContentAllowList":
        if not hasattr(cls, "instance"):
            obj = cls.instance = super(ContentAllowList, cls).__new__(cls)
            obj._allow_list: list[dict] = []

        return cls.instance

    def is_allowed(self, association_properties: dict) -> bool:
        for allow_list_item in self._allow_list:
            if all(
                [
                    association_properties.get(key) == value
                    for key, value in allow_list_item.items()
                ]
            ):
                return True

        return False

    def load(self, response_json: dict):
        self._allow_list = response_json["associationPropertyAllowList"]



from functools import wraps
from typing import Optional

from opentelemetry.semconv.ai import SpanAttributes, TraceloopSpanKindValues

from traceloop.sdk.tracing import get_tracer, set_workflow_name
from traceloop.sdk.tracing.tracing import TracerWrapper
from traceloop.sdk.utils import camel_to_snake


def task(
    name: Optional[str] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    if method_name is None:
        return task_method(name=name, tlp_span_kind=tlp_span_kind)
    else:
        return task_class(
            name=name, method_name=method_name, tlp_span_kind=tlp_span_kind
        )


def task_method(
    name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    def decorate(fn):
        @wraps(fn)
        def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return fn(*args, **kwargs)

            span_name = (
                f"{name}.{tlp_span_kind.value}"
                if name
                else f"{fn.__name__}.{tlp_span_kind.value}"
            )
            with get_tracer() as tracer, tracer.start_as_current_span(
                span_name
            ) as span:
                span.set_attribute(
                    SpanAttributes.TRACELOOP_SPAN_KIND, tlp_span_kind.value
                )
                span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)
                return fn(*args, **kwargs)

        return wrap

    return decorate


def task_class(
    name: Optional[str],
    method_name: str,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    def decorator(cls):
        task_name = name if name else camel_to_snake(cls.__name__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            task_method(name=task_name, tlp_span_kind=tlp_span_kind)(method),
        )
        return cls

    return decorator


def workflow(
    name: Optional[str] = None,
    method_name: Optional[str] = None,
    correlation_id: Optional[str] = None,
):
    if method_name is None:
        return workflow_method(name=name, correlation_id=correlation_id)
    else:
        return workflow_class(
            name=name, method_name=method_name, correlation_id=correlation_id
        )


def workflow_method(name: Optional[str] = None, correlation_id: Optional[str] = None):
    def decorate(fn):
        @wraps(fn)
        def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return fn(*args, **kwargs)

            workflow_name = name or fn.__name__
            set_workflow_name(workflow_name)
            span_name = f"{workflow_name}.workflow"

            with (
                get_tracer(flush_on_exit=True) as tracer,
                tracer.start_as_current_span(span_name) as span,
            ):
                span.set_attribute(
                    SpanAttributes.TRACELOOP_SPAN_KIND,
                    TraceloopSpanKindValues.WORKFLOW.value,
                )
                span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)

                return fn(*args, **kwargs)

        return wrap

    return decorate


def workflow_class(
    name: Optional[str], method_name: str, correlation_id: Optional[str] = None
):
    def decorator(cls):
        workflow_name = name if name else camel_to_snake(cls.__name__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            workflow_method(name=workflow_name, correlation_id=correlation_id)(method),
        )
        return cls

    return decorator


def agent(name: Optional[str] = None, method_name: Optional[str] = None):
    return task(
        name=name, method_name=method_name, tlp_span_kind=TraceloopSpanKindValues.AGENT
    )


def tool(name: Optional[str] = None, method_name: Optional[str] = None):
    return task(
        name=name, method_name=method_name, tlp_span_kind=TraceloopSpanKindValues.TOOL
    )


# Async Decorators
def atask(
    name: Optional[str] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    if method_name is None:
        return atask_method(name=name, tlp_span_kind=tlp_span_kind)
    else:
        return atask_class(
            name=name, method_name=method_name, tlp_span_kind=tlp_span_kind
        )


def atask_method(
    name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    def decorate(fn):
        @wraps(fn)
        async def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return await fn(*args, **kwargs)

            span_name = (
                f"{name}.{tlp_span_kind.value}"
                if name
                else f"{fn.__name__}.{tlp_span_kind.value}"
            )
            with (
                get_tracer() as tracer,
                tracer.start_as_current_span(span_name) as span,
            ):
                span.set_attribute(
                    SpanAttributes.TRACELOOP_SPAN_KIND, tlp_span_kind.value
                )
                span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)
                return await fn(*args, **kwargs)

        return wrap

    return decorate


def atask_class(
    name: Optional[str],
    method_name: str,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    def decorator(cls):
        task_name = name if name else camel_to_snake(cls.__name__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            atask_method(name=task_name, tlp_span_kind=tlp_span_kind)(method),
        )
        return cls

    return decorator


def aworkflow(
    name: Optional[str] = None,
    method_name: Optional[str] = None,
    correlation_id: Optional[str] = None,
):
    if method_name is None:
        return aworkflow_method(name=name, correlation_id=correlation_id)
    else:
        return aworkflow_class(
            name=name, method_name=method_name, correlation_id=correlation_id
        )


def aworkflow_method(name: Optional[str] = None, correlation_id: Optional[str] = None):
    def decorate(fn):
        @wraps(fn)
        async def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return await fn(*args, **kwargs)

            workflow_name = name or fn.__name__
            set_workflow_name(workflow_name)
            span_name = f"{workflow_name}.workflow"

            with (
                get_tracer(flush_on_exit=True) as tracer,
                tracer.start_as_current_span(span_name) as span,
            ):
                span.set_attribute(
                    SpanAttributes.TRACELOOP_SPAN_KIND,
                    TraceloopSpanKindValues.WORKFLOW.value,
                )
                span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)

                if correlation_id:
                    span.set_attribute(
                        SpanAttributes.TRACELOOP_CORRELATION_ID, correlation_id
                    )
                return await fn(*args, **kwargs)

        return wrap

    return decorate


def aworkflow_class(
    name: Optional[str], method_name: str, correlation_id: Optional[str] = None
):
    def decorator(cls):
        workflow_name = name if name else camel_to_snake(cls.__name__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            aworkflow_method(name=workflow_name, correlation_id=correlation_id)(method),
        )
        return cls

    return decorator


def aagent(name: Optional[str] = None, method_name: Optional[str] = None):
    return atask(
        name=name, method_name=method_name, tlp_span_kind=TraceloopSpanKindValues.AGENT
    )


def atool(name: Optional[str] = None, method_name: Optional[str] = None):
    return atask(
        name=name, method_name=method_name, tlp_span_kind=TraceloopSpanKindValues.TOOL
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sdk',
 'sdk.config',
 'sdk.decorators',
 'sdk.prompts',
 'sdk.tracing',
 'sdk.utils']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.6,<0.5.0',
 'deprecated>=1.2.14,<2.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'opentelemetry-api>=1.20.0,<2.0.0',
 'opentelemetry-exporter-otlp-proto-grpc>=1.20.0,<2.0.0',
 'opentelemetry-exporter-otlp-proto-http>=1.20.0,<2.0.0',
 'opentelemetry-instrumentation-anthropic>=0.0.12,<0.0.13',
 'opentelemetry-instrumentation-chromadb==0.0.6',
 'opentelemetry-instrumentation-cohere>=0.0.8,<0.0.9',
 'opentelemetry-instrumentation-langchain==0.0.7',
 'opentelemetry-instrumentation-openai>=0.0.16,<0.0.17',
 'opentelemetry-instrumentation-pinecone>=0.0.10,<0.0.11',
 'opentelemetry-instrumentation-pymysql>=0.41b0,<0.42',
 'opentelemetry-instrumentation-requests>=0.41b0,<0.42',
 'opentelemetry-instrumentation-transformers==0.0.4',
 'opentelemetry-instrumentation-urllib3>=0.41b0,<0.42',
 'opentelemetry-sdk>=1.20.0,<2.0.0',
 'opentelemetry-semantic-conventions-ai>=0.0.7,<0.0.8',
 'pydantic>=1.10.12,<2.0.0',
 'tenacity>=8.2.3,<9.0.0']

setup_kwargs = {
    'name': 'traceloop-sdk',
    'version': '0.0.70',
    'description': 'Traceloop Software Development Kit (SDK) for Python',
    'long_description': '# traceloop-sdk\n\nTraceloop’s Python SDK allows you to easily start monitoring and debugging your LLM execution. Tracing is done in a non-intrusive way, built on top of OpenTelemetry. You can choose to export the traces to Traceloop, or to your existing observability stack.\n\n```python\nTraceloop.init(app_name="joke_generation_service")\n\n@workflow(name="joke_creation")\ndef create_joke():\n    completion = openai.ChatCompletion.create(\n        model="gpt-3.5-turbo",\n        messages=[{"role": "user", "content": "Tell me a joke about opentelemetry"}],\n    )\n\n    return completion.choices[0].message.content\n```\n',
    'author': 'Gal Kleinman',
    'author_email': 'gal@traceloop.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/traceloop/openllmetry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4',
}


setup(**setup_kwargs)



import os
import sys
from deprecated import deprecated
import requests

from typing import Optional
from colorama import Fore
from opentelemetry.sdk.trace.export import SpanExporter
from opentelemetry.util.re import parse_env_headers

from traceloop.sdk.config import (
    is_content_tracing_enabled,
    is_tracing_enabled,
)
from traceloop.sdk.fetcher import Fetcher
from traceloop.sdk.prompts.client import PromptRegistryClient
from traceloop.sdk.tracing.content_allow_list import ContentAllowList
from traceloop.sdk.tracing.tracing import (
    TracerWrapper,
    set_association_properties,
    set_correlation_id,
)


class Traceloop:
    __tracer_wrapper: TracerWrapper

    @staticmethod
    def init(
        app_name: Optional[str] = sys.argv[0],
        api_endpoint: str = "https://api.traceloop.com",
        api_key: str = None,
        headers: dict[str, str] = {},
        disable_batch=False,
        exporter: SpanExporter = None,
        traceloop_sync_enabled: bool = True,
    ) -> None:
        api_endpoint = os.getenv("TRACELOOP_BASE_URL") or api_endpoint

        if (
            traceloop_sync_enabled
            and api_endpoint.find("traceloop.com") != -1
            and not exporter
        ):
            Fetcher(base_url=api_endpoint).run()
            print(
                Fore.GREEN + "Tracloop syncing configuration and prompts" + Fore.RESET
            )

        if not is_tracing_enabled():
            print(Fore.YELLOW + "Tracing is disabled" + Fore.RESET)
            return

        enable_content_tracing = is_content_tracing_enabled()

        if exporter:
            print(Fore.GREEN + "Traceloop exporting traces to a custom exporter")

        api_key = os.getenv("TRACELOOP_API_KEY") or api_key
        headers = os.getenv("TRACELOOP_HEADERS") or headers

        if isinstance(headers, str):
            headers = parse_env_headers(headers)

        # auto-create a dashboard on Traceloop if no export endpoint is provided
        if not exporter and api_endpoint == "https://api.traceloop.com" and not api_key:
            headers = None  # disable headers if we're auto-creating a dashboard
            if os.path.exists("/tmp/traceloop_key.txt") and os.path.exists(
                "/tmp/traceloop_url.txt"
            ):
                api_key = open("/tmp/traceloop_key.txt").read()
                access_url = open("/tmp/traceloop_url.txt").read()
            else:
                print(
                    Fore.YELLOW
                    + "No Traceloop API key provided, auto-creating a dashboard on Traceloop",
                )
                res = requests.post(
                    "https://app.traceloop.com/api/registration/auto-create"
                ).json()
                access_url = f"https://app.traceloop.com/trace?skt={res['uiAccessKey']}"
                api_key = res["apiKey"]
                print(Fore.YELLOW + "TRACELOOP_API_KEY=", api_key)
                open("/tmp/traceloop_key.txt", "w").write(api_key)
                open("/tmp/traceloop_url.txt", "w").write(access_url)
            print(
                Fore.GREEN + f"\nGo to {access_url} to see a live dashboard\n",
            )

        if headers:
            print(
                Fore.GREEN
                + f"Traceloop exporting traces to {api_endpoint}, authenticating with custom headers"
            )

        if api_key and not headers:
            print(
                Fore.GREEN
                + f"Traceloop exporting traces to {api_endpoint} authenticating with bearer token"
            )
            headers = {
                "Authorization": f"Bearer {api_key}",
            }

        print(Fore.RESET)

        TracerWrapper.set_static_params(
            app_name, enable_content_tracing, api_endpoint, headers
        )
        Traceloop.__tracer_wrapper = TracerWrapper(
            disable_batch=disable_batch, exporter=exporter
        )

    @staticmethod
    @deprecated(version="0.0.62", reason="Use set_association_properties instead")
    def set_correlation_id(correlation_id: str) -> None:
        set_correlation_id(correlation_id)

    def set_association_properties(properties: dict) -> None:
        set_association_properties(properties)
